"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ha_t():
    """
    Test case:  Ordered list newline atx heading text
    was:        test_list_blocks_256g
    """

    # Arrange
    source_markdown = """1.
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[end-olist:::True]",
        "[atx(2,1):1:0:]",
        "[text(2,3):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<ol>
<li></li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_i2_ha_t():
    """
    Test case:  Ordered list newline indent of 2 atx heading text
    was:        test_list_blocks_256gxa
    """

    # Arrange
    source_markdown = """1.
  # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[end-olist:::True]",
        "[atx(2,3):1:0:  ]",
        "[text(2,5):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<ol>
<li></li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_i3_ha_t():
    """
    Test case:  Ordered list newline indent of 3 atx heading text
    was:        test_list_blocks_256gxb
    """

    # Arrange
    source_markdown = """1.
   # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[BLANK(1,3):]",
        "[atx(2,4):1:0:]",
        "[text(2,6):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<h1>foo</h1>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ha_t():
    """
    Test case:  Ordered list text newline atx heading text
    was:        test_list_blocks_256ga
    """

    # Arrange
    source_markdown = """1. abc
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[atx(2,1):1:0:]",
        "[text(2,3):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<ol>
<li>abc</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_i2_ha_t():
    """
    Test case:  Ordered list text newline indent of 2 atx heading text
    was:        test_list_blocks_256gaa
    """

    # Arrange
    source_markdown = """1. abc
  # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[atx(2,3):1:0:  ]",
        "[text(2,5):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<ol>
<li>abc</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_i3_ha_t():
    """
    Test case:  Ordered list text newline indent of 3 atx heading text
    was:        test_list_blocks_256gab
    """

    # Arrange
    source_markdown = """1. abc
   # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::False]",
        "[atx(2,4):1:0:]",
        "[text(2,6):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>abc
<h1>foo</h1>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_ol_nl_ha_t():
    """
    Test case:  Ordered list x2 newline atx heading text
    """

    # Arrange
    source_markdown = """1. 1.
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   ]",
        "[BLANK(1,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(2,1):1:0:]",
        "[text(2,3):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_ol_t_nl_ha_t():
    """
    Test case:  Ordered list x2 text newline atx heading text
    was:        test_list_blocks_256gb
    """

    # Arrange
    source_markdown = """1. 1. abc
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   ]",
        "[para(1,7):]",
        "[text(1,7):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(2,1):1:0:]",
        "[text(2,3):foo: ]",
        "[end-atx::]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc</li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ol_nl_ha_t():
    """
    Test case:  Ordered list newline ordered list new line atx heading text
    """

    # Arrange
    source_markdown = """1.
   1.
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(3,1):1:0:]",
        "[text(3,3):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ol_nl_ha_t():
    """
    Test case:  Ordered list text newline ordered list new line atx heading text
    """

    # Arrange
    source_markdown = """1. abc
   1.
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\n1.::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[atx(3,1):1:0:]",
        "[text(3,3):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ul_t_nl_ul_nl_ha_t():
    """
    Test case:  Unordered list text newline unordered list new line atx heading text
    """

    # Arrange
    source_markdown = """- abc
  -
# foo
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[end-ulist:::True]",
        "[atx(3,1):1:0:]",
        "[text(3,3):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
</li>
</ul>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ul_t_nl_ulb_nl_ha_t():
    """
    Test case:  Unordered list text newline unordered list (b) new line atx heading text
    """

    # Arrange
    source_markdown = """- abc
  *
# foo
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[atx(3,1):1:0:]",
        "[text(3,3):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*</li>
</ul>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ol_t_nl_ha_t():
    """
    Test case:  Ordered list newline ordered list text new line atx heading text
    was:        test_list_blocks_256gc
    """

    # Arrange
    source_markdown = """1.
   1. abc
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(3,1):1:0:]",
        "[text(3,3):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc</li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ol_t_nl_ha_t():
    """
    Test case:  Ordered list text newline ordered list text new line atx heading text
    """

    # Arrange
    source_markdown = """1. abc
   1. abc
# foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(3,1):1:0:]",
        "[text(3,3):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>abc</li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ol_nl_i2_ha_t():
    """
    Test case:  Ordered list newline ordered list new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """1.
   1.
  # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(3,3):1:0:  ]",
        "[text(3,5):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ol_nl_i2_ha_t():
    """
    Test case:  Ordered list text newline ordered list new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """1. abc
   1.
  # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\n1.::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[atx(3,3):1:0:  ]",
        "[text(3,5):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ul_t_nl_ul_nl_i1_ha_t():
    """
    Test case:  Unordered list text newline unordered list new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """- abc
  -
 # foo
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[end-ulist:::True]",
        "[atx(3,2):1:0: ]",
        "[text(3,4):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
</li>
</ul>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ul_t_nl_ulb_nl_i1_ha_t():
    """
    Test case:  Unordered list text newline unordered list (b) new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """- abc
  *
 # foo
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[atx(3,2):1:0: ]",
        "[text(3,4):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*</li>
</ul>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ol_t_nl_i2_ha_t():
    """
    Test case:  Ordered list newline ordered list text new line indent of 2 atx heading text
    was:        test_list_blocks_256gd
    """

    # Arrange
    source_markdown = """1.
   1. abc
  # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(3,3):1:0:  ]",
        "[text(3,5):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc</li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ol_t_nl_i2_ha_t():
    """
    Test case:  Ordered list text newline ordered list text new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """1. abc
   1. abc
  # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[atx(3,3):1:0:  ]",
        "[text(3,5):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>abc</li>
</ol>
</li>
</ol>
<h1>foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ol_nl_i3_ha_t():
    """
    Test case:  Ordered list newline ordered list new line indent of 3 atx heading text
    """

    # Arrange
    source_markdown = """1.
   1.
   # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[atx(3,4):1:0:]",
        "[text(3,6):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
<h1>foo</h1>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ol_nl_i3_ha_t():
    """
    Test case:  Ordered list text newline ordered list new line indent of 3 atx heading text
    """

    # Arrange
    source_markdown = """1. abc
   1.
   # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\n1.::\n]",
        "[end-para:::False]",
        "[atx(3,4):1:0:]",
        "[text(3,6):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>abc
1.
<h1>foo</h1>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ul_t_nl_ul_nl_i2_ha_t():
    """
    Test case:  Unordered list text newline unordered list new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """- abc
  -
  # foo
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[atx(3,3):1:0:]",
        "[text(3,5):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
<h1>foo</h1>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ul_t_nl_ulb_nl_i2_ha_t():
    """
    Test case:  Unordered list text newline unordered list (b) new line indent of 2 atx heading text
    """

    # Arrange
    source_markdown = """- abc
  *
  # foo
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[para(1,3):\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[end-para:::False]",
        "[atx(3,3):1:0:]",
        "[text(3,5):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>abc
*
<h1>foo</h1>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_nl_ol_t_nl_i3_ha_t():
    """
    Test case:  Ordered list newline ordered list text new line indent of 3 atx heading text
    was:        test_list_blocks_256ge
    """

    # Arrange
    source_markdown = """1.
   1. abc
   # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   :   ]",
        "[para(2,7):]",
        "[text(2,7):abc:]",
        "[end-para:::False]",
        "[end-olist:::True]",
        "[atx(3,4):1:0:]",
        "[text(3,6):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc</li>
</ol>
<h1>foo</h1>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_ha_ol_t_nl_ol_t_nl_i3_ha_t():
    """
    Test case:  Ordered list text newline ordered list text new line indent of 3 atx heading text
    """

    # Arrange
    source_markdown = """1. abc
   1. abc
   # foo
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   :   ]",
        "[para(2,7):]",
        "[text(2,7):abc:]",
        "[end-para:::False]",
        "[end-olist:::True]",
        "[atx(3,4):1:0:]",
        "[text(3,6):foo: ]",
        "[end-atx::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>abc</li>
</ol>
<h1>foo</h1>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide for a container element that can be added to markdown parsing stream.
"""
import logging

from pymarkdown.markdown_token import MarkdownToken, MarkdownTokenClass
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-many-arguments
class ContainerMarkdownToken(MarkdownToken):
    """
    Class to provide for a container element that can be added to markdown parsing stream.
    """

    def __init__(
        self,
        token_name,
        extra_data,
        line_number=0,
        column_number=0,
        position_marker=None,
    ):
        MarkdownToken.__init__(
            self,
            token_name,
            MarkdownTokenClass.CONTAINER_BLOCK,
            extra_data,
            line_number=line_number,
            column_number=column_number,
            position_marker=position_marker,
            requires_end_token=True,
        )


# pylint: enable=too-many-arguments


class NewListItemMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of the new list item element.
    """

    def __init__(
        self, indent_level, position_marker, extracted_whitespace, list_start_content
    ):
        self.__indent_level, self.__extracted_whitespace, self.__list_start_content = (
            indent_level,
            extracted_whitespace,
            list_start_content,
        )
        ContainerMarkdownToken.__init__(
            self,
            MarkdownToken._token_new_list_item,
            MarkdownToken.extra_data_separator.join(
                [str(indent_level), extracted_whitespace, list_start_content]
            ),
            position_marker=position_marker,
        )

    @property
    def extracted_whitespace(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def indent_level(self):
        """
        Returns the indent level to apply for the list element.
        """
        return self.__indent_level

    @property
    def list_start_content(self):
        """
        Returns the content used to start this list element.
        """
        return self.__list_start_content


# pylint: disable=too-many-instance-attributes
class OrderedListStartMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of the ordered list start element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        list_start_sequence,
        list_start_content,
        indent_level,
        extracted_whitespace,
        position_marker,
    ):
        (
            self.__list_start_sequence,
            self.__list_start_content,
            self.__indent_level,
            self.__extracted_whitespace,
            self.__leading_spaces,
            self.is_loose,
            self.leading_spaces_index,
        ) = (
            list_start_sequence,
            list_start_content,
            indent_level,
            extracted_whitespace,
            None,
            True,
            0,
        )
        ContainerMarkdownToken.__init__(
            self,
            MarkdownToken._token_ordered_list_start,
            "",
            position_marker=position_marker,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments

    @property
    def extracted_whitespace(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def indent_level(self):
        """
        Returns the indent level to apply for the list element.
        """
        return self.__indent_level

    @property
    def list_start_sequence(self):
        """
        Returns the sequence used to start this list element.
        """
        return self.__list_start_sequence

    @property
    def list_start_content(self):
        """
        Returns the content used to start this list element.
        """
        return self.__list_start_content

    @property
    def leading_spaces(self):
        """
        Returns the leading spaces that occur before the list element.
        """
        return self.__leading_spaces

    def adjust_for_new_list_item(self, new_list_item_token):
        """
        Adjust this token for a new list item (uses copy to keep track).
        """
        assert new_list_item_token and new_list_item_token.is_new_list_item

        self.__indent_level, self.__extracted_whitespace = (
            new_list_item_token.indent_level,
            new_list_item_token.extracted_whitespace,
        )
        self.__compose_extra_data_field()

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        item_list = [
            self.__list_start_sequence,
            self.__list_start_content,
            str(self.__indent_level),
            self.__extracted_whitespace,
        ]
        if self.__leading_spaces is not None:
            item_list.append(self.__leading_spaces)
        self._set_extra_data(MarkdownToken.extra_data_separator.join(item_list))

    def add_leading_spaces(self, ws_add):
        """
        Add any leading spaces to the token, separating them with line feeds.
        """
        self.__leading_spaces = (
            ws_add
            if self.__leading_spaces is None
            else f"{self.__leading_spaces}{ParserHelper.newline_character}{ws_add}"
        )
        self.__compose_extra_data_field()


# pylint: enable=too-many-instance-attributes


class UnorderedListStartMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of the unordered list start element.
    """

    def __init__(
        self, list_start_sequence, indent_level, extracted_whitespace, position_marker
    ):
        (
            self.__list_start_sequence,
            self.__indent_level,
            self.__extracted_whitespace,
            self.__leading_spaces,
            self.is_loose,
            self.leading_spaces_index,
        ) = (list_start_sequence, indent_level, extracted_whitespace, None, True, 0)
        ContainerMarkdownToken.__init__(
            self,
            MarkdownToken._token_unordered_list_start,
            "",
            position_marker=position_marker,
        )
        self.__compose_extra_data_field()

    @property
    def extracted_whitespace(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def indent_level(self):
        """
        Returns the indent level to apply for the list element.
        """
        return self.__indent_level

    @property
    def list_start_sequence(self):
        """
        Returns the sequence used to start this list element.
        """
        return self.__list_start_sequence

    @property
    def leading_spaces(self):
        """
        Returns the leading spaces that occur before the list element.
        """
        return self.__leading_spaces

    def adjust_for_new_list_item(self, new_list_item_token):
        """
        Adjust this token for a new list item (uses copy to keep track).
        """
        assert new_list_item_token and new_list_item_token.is_new_list_item

        self.__indent_level, self.__extracted_whitespace = (
            new_list_item_token.indent_level,
            new_list_item_token.extracted_whitespace,
        )
        self.__compose_extra_data_field()

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        item_list = [
            self.__list_start_sequence,
            "",
            str(self.__indent_level),
            self.__extracted_whitespace,
        ]
        if self.__leading_spaces is not None:
            item_list.append(self.__leading_spaces)
        self._set_extra_data(MarkdownToken.extra_data_separator.join(item_list))

    def add_leading_spaces(self, ws_add):
        """
        Add any leading spaces to the token, separating them with line feeds.
        """
        self.__leading_spaces = (
            ws_add
            if self.__leading_spaces is None
            else f"{self.__leading_spaces}{ParserHelper.newline_character}{ws_add}"
        )
        self.__compose_extra_data_field()


class BlockQuoteMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of the block quote element.
    """

    def __init__(self, extracted_whitespace, position_marker):
        self.__extracted_whitespace, self.__leading_spaces, self.leading_text_index = (
            extracted_whitespace,
            "",
            0,
        )
        ContainerMarkdownToken.__init__(
            self,
            MarkdownToken._token_block_quote,
            "",
            position_marker=position_marker,
        )
        self.__compose_extra_data_field()

    @property
    def extracted_whitespace(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def leading_spaces(self):
        """
        Returns any leading spaces that preface the block quote.
        """
        return self.__leading_spaces

    def add_leading_spaces(self, leading_spaces_to_add):
        """
        Add any leading spaces to the token, separating them with line feeds.
        """
        POGGER.debug(
            "__leading_spaces>>:$:<<",
            self.__leading_spaces,
        )
        POGGER.debug("add_leading_spaces>>:$:<<", leading_spaces_to_add)
        self.__leading_spaces = (
            f"{self.__leading_spaces}{ParserHelper.newline_character}{leading_spaces_to_add}"
            if self.__leading_spaces
            else leading_spaces_to_add
        )
        POGGER.debug(
            "__leading_spaces>>:$:<<",
            self.__leading_spaces,
        )
        self.__compose_extra_data_field()

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        item_list = [self.__extracted_whitespace, self.__leading_spaces]
        self._set_extra_data(MarkdownToken.extra_data_separator.join(item_list))

    def calculate_next_leading_space_part(self, increment_index=True):
        """
        Calculate the next leading space based on the leading_text_index,
        optonally incrementing it as well.
        """
        split_leading_spaces = self.leading_spaces.split(ParserHelper.newline_character)
        leading_text = split_leading_spaces[self.leading_text_index]
        if increment_index:
            self.leading_text_index += 1
        return leading_text



"""
Module to provide a tokenization of a markdown-encoded string.
"""
import logging
import os

from pymarkdown.application_properties import ApplicationProperties
from pymarkdown.bad_tokenization_error import BadTokenizationError
from pymarkdown.coalesce_processor import CoalesceProcessor
from pymarkdown.container_block_processor import ContainerBlockProcessor
from pymarkdown.extensions.front_matter_markdown_token import FrontMatterExtension
from pymarkdown.extensions.pragma_token import PragmaToken
from pymarkdown.html_helper import HtmlHelper
from pymarkdown.inline_helper import InlineHelper
from pymarkdown.inline_processor import InlineProcessor
from pymarkdown.leaf_block_processor import LeafBlockProcessor
from pymarkdown.leaf_markdown_token import BlankLineMarkdownToken
from pymarkdown.link_helper import LinkHelper
from pymarkdown.link_reference_definition_helper import LinkReferenceDefinitionHelper
from pymarkdown.parser_helper import ParserHelper, ParserState, PositionMarker
from pymarkdown.parser_logger import ParserLogger
from pymarkdown.source_providers import InMemorySourceProvider
from pymarkdown.stack_token import DocumentStackToken, ParagraphStackToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class TokenizedMarkdown:
    """
    Class to provide a tokenization of a markdown-encoded string.
    """

    def __init__(self, resource_path=None):
        """
        Initializes a new instance of the TokenizedMarkdown class.
        """

        (
            self.tokenized_document,
            self.stack,
            self.source_provider,
            self.__properties,
            self.__front_matter_enabled,
        ) = (None, None, None, ApplicationProperties(), None)

        if not resource_path:
            resource_path = os.path.join(os.path.split(__file__)[0], "resources")
        InlineHelper.initialize(resource_path)

    def apply_configuration(self, application_properties):
        """
        Apply any configuration map.
        """
        self.__properties = application_properties
        self.__front_matter_enabled = self.__properties.get_boolean_property(
            "extensions.front-matter.enabled", default_value=False
        )

    def transform_from_provider(self, source_provider):
        """
        Transform the data from the source provider into a Markdown token stream.
        """
        self.source_provider = source_provider
        return self.__transform()

    def transform(self, your_text_string, show_debug=False):
        """
        Transform a text string in a Markdown format into a Markdown token stream.
        This function should only be used as a simplified manner of accessing the
        functionality for the purposes of testing.
        """

        logging.getLogger().setLevel(logging.DEBUG if show_debug else logging.WARNING)
        ParserLogger.sync_on_next_call()
        self.source_provider = InMemorySourceProvider(your_text_string)
        return self.__transform()

    def __transform(self):
        """
        Transform a markdown-encoded string into an array of tokens.
        """
        try:
            self.tokenized_document, self.stack = None, []

            InlineProcessor.initialize()
            LinkHelper.initialize()

            POGGER.debug("\n\n>>>>>>>parse_blocks_pass>>>>>>")
            first_pass_results = self.__parse_blocks_pass()

            POGGER.debug("\n\n>>>>>>>coalesce_text_blocks>>>>>>")
            coalesced_results = CoalesceProcessor.coalesce_text_blocks(
                first_pass_results
            )

            POGGER.debug("\n\n>>>>>>>parse_inline>>>>>>")
            final_pass_results = InlineProcessor.parse_inline(coalesced_results)

            POGGER.debug("\n\n>>>>>>>final_pass_results>>>>>>")
            return final_pass_results
        except Exception as this_exception:
            raise BadTokenizationError(
                "An unhandled error occurred processing the document."
            ) from this_exception

    # pylint: disable=too-many-statements,too-many-locals,too-many-branches
    def __parse_blocks_pass(self):
        """
        The first pass at the tokens is to deal with blocks.
        """

        self.stack = [DocumentStackToken()]

        self.tokenized_document = []
        token_to_use = self.source_provider.get_next_line()
        did_start_close = False
        did_started_close = False
        requeue = []
        ignore_link_definition_start = False
        POGGER.debug("---$---", token_to_use)
        POGGER.debug("---")
        pragma_lines = {}
        line_number = 1
        try:
            token_to_use, line_number, requeue = self.__process_header_if_present(
                token_to_use, line_number, requeue
            )
            did_start_close = token_to_use is None
            keep_on_going = True
            while keep_on_going:
                POGGER.debug("next-line>>$", token_to_use)
                POGGER.debug("stack>>$", self.stack)
                POGGER.debug("current_block>>$", self.stack[-1])
                POGGER.debug("line_number>>$", line_number)
                POGGER.debug("---")

                position_marker = PositionMarker(line_number, 0, token_to_use)
                parser_state = ParserState(
                    self.stack,
                    self.tokenized_document,
                    TokenizedMarkdown.__close_open_blocks,
                    self.__handle_blank_line,
                )
                if did_start_close:
                    POGGER.debug("\n\ncleanup")

                    was_link_definition_started_before_close = self.stack[
                        -1
                    ].was_link_definition_started

                    did_started_close = True
                    (
                        tokens_from_line,
                        requeue_line_info,
                    ) = TokenizedMarkdown.__close_open_blocks(
                        parser_state,
                        self.tokenized_document,
                        include_block_quotes=True,
                        include_lists=True,
                        caller_can_handle_requeue=True,
                        was_forced=True,
                    )
                    if tokens_from_line and not self.tokenized_document:
                        self.tokenized_document.extend(tokens_from_line)

                    if not (requeue_line_info and requeue_line_info.lines_to_requeue):
                        keep_on_going = False
                    else:
                        assert was_link_definition_started_before_close
                        assert not requeue_line_info.lines_to_requeue[0]

                        del requeue_line_info.lines_to_requeue[0]
                        line_number -= 1

                        did_start_close = False
                        tokens_from_line = None
                else:
                    POGGER.debug(">>>>$", self.tokenized_document)

                    if not token_to_use or not token_to_use.strip():
                        POGGER.debug("call __parse_blocks_pass>>handle_blank_line")
                        (
                            tokens_from_line,
                            requeue_line_info,
                        ) = self.__handle_blank_line(
                            parser_state,
                            token_to_use,
                            from_main_transform=True,
                            position_marker=position_marker,
                        )
                    else:
                        POGGER.debug("\n\nnormal lines")
                        (
                            tokens_from_line,
                            _,
                            requeue_line_info,
                        ) = ContainerBlockProcessor.parse_line_for_container_blocks(
                            parser_state,
                            position_marker,
                            ignore_link_definition_start,
                            pragma_lines=pragma_lines,
                        )

                    POGGER.debug("<<<<$", self.tokenized_document)

                if keep_on_going:
                    line_number, ignore_link_definition_start = TokenizedMarkdown.__xx(
                        line_number, requeue_line_info, requeue
                    )

                    POGGER.debug(
                        "---\nbefore>>$",
                        self.tokenized_document,
                    )
                    POGGER.debug("before>>$", tokens_from_line)
                    if tokens_from_line:
                        self.tokenized_document.extend(tokens_from_line)
                    POGGER.debug(
                        "after>>$",
                        self.tokenized_document,
                    )
                    if requeue:
                        POGGER.debug("requeue>>$", requeue)
                    POGGER.debug("---")

                    (
                        token_to_use,
                        did_start_close,
                        did_started_close,
                    ) = self.__determine_next_token_process(
                        requeue, did_start_close, did_started_close
                    )
        except AssertionError as this_exception:
            error_message = f"A project assertion failed on line {str(line_number)} of the current document."
            raise BadTokenizationError(error_message) from this_exception

        if pragma_lines:
            self.tokenized_document.append(PragmaToken(pragma_lines))
        return self.tokenized_document

    # pylint: enable=too-many-statements,too-many-locals,too-many-branches

    @staticmethod
    def __xx(line_number, requeue_line_info, requeue):

        if requeue_line_info and requeue_line_info.lines_to_requeue:
            number_of_lines_to_requeue = len(requeue_line_info.lines_to_requeue)
            POGGER.debug("\n\n---lines_to_requeue>>$", number_of_lines_to_requeue)
            line_number -= number_of_lines_to_requeue - 1

            for i in requeue_line_info.lines_to_requeue:
                requeue.insert(0, i)
            ignore_link_definition_start = requeue_line_info.force_ignore_first_as_lrd
        else:
            ignore_link_definition_start = False
            line_number += 1
        POGGER.debug("line_number>>$\n---", line_number)

        return line_number, ignore_link_definition_start

    def __determine_next_token_process(
        self, requeue, did_start_close, did_started_close
    ):
        """
        For the parse_blocks_pass function, determine the next token to parse.
        """

        token_to_use = None
        if requeue:
            POGGER.debug(">>Requeues present")
            token_to_use = requeue[0]
            del requeue[0]
            POGGER.debug(">>Requeue>>$", token_to_use)
            POGGER.debug(">>Requeues left>>$", requeue)
        elif did_started_close:
            did_start_close = True
        else:
            token_to_use = self.source_provider.get_next_line()
            if token_to_use is None:
                did_start_close = True

        return token_to_use, did_start_close, did_started_close

    # pylint: disable=too-many-arguments,too-many-locals,too-many-statements, too-many-branches
    @staticmethod
    def __close_open_blocks(  # noqa: C901
        parser_state,
        destination_array=None,
        only_these_blocks=None,
        include_block_quotes=False,
        include_lists=False,
        until_this_index=-1,
        caller_can_handle_requeue=False,
        was_forced=False,
    ):
        """
        Close any open blocks that are currently on the stack.
        """

        new_tokens = []
        requeue_line_info = None
        if destination_array:
            new_tokens = destination_array

        POGGER.debug("cob-start>>$", parser_state.token_stack)
        POGGER.debug(
            "cob-start>>$",
            parser_state.token_document,
        )
        if destination_array:
            POGGER.debug(
                "cob-destination_array>>$",
                destination_array,
            )
        if only_these_blocks:
            POGGER.debug("cob-only_these_blocks>>$", only_these_blocks)
        if include_block_quotes:
            POGGER.debug("cob-include_block_quotes>>$", include_block_quotes)
        if include_lists:
            POGGER.debug("cob-include_lists>>$", include_lists)
        if until_this_index != -1:
            POGGER.debug("cob-until_this_index>>$", until_this_index)
        if caller_can_handle_requeue:
            POGGER.debug("cob-caller_can_handle_requeue>>$", caller_can_handle_requeue)
        if was_forced:
            POGGER.debug("cob-was_forced>>$", was_forced)
        while not parser_state.token_stack[-1].is_document:

            was_close_forced = was_forced
            POGGER.debug("cob>>$", parser_state.token_stack)
            if only_these_blocks:
                POGGER.debug("cob-only-type>>$", only_these_blocks)
                POGGER.debug("cob-only-type>>$", type(parser_state.token_stack[-1]))
                # pylint: disable=unidiomatic-typecheck
                if type(parser_state.token_stack[-1]) not in only_these_blocks:
                    POGGER.debug("cob>>not in only")
                    break
                # pylint: enable=unidiomatic-typecheck
            if not include_block_quotes and parser_state.token_stack[-1].is_block_quote:
                POGGER.debug("cob>>not block quotes")
                break
            if not include_lists and parser_state.token_stack[-1].is_list:
                POGGER.debug("cob>>not lists")
                break
            if until_this_index != -1:
                token_stack_size = len(parser_state.token_stack)
                POGGER.debug(
                    "NOT ME!!!!$<<$<<",
                    until_this_index,
                    token_stack_size,
                )
                if until_this_index >= token_stack_size:
                    break
                was_close_forced = True

            if parser_state.token_stack[-1].was_link_definition_started:
                POGGER.debug(
                    "cob->process_link_reference_definition>>stopping link definition"
                )
                empty_position_marker = PositionMarker(-1, 0, "")
                (
                    outer_processed,
                    did_complete_lrd,
                    did_pause_lrd,
                    requeue_line_info,
                    adjusted_tokens,
                ) = LinkReferenceDefinitionHelper.process_link_reference_definition(
                    parser_state, empty_position_marker, "", "", "", 0, 0
                )
                POGGER.debug("BOOOM")
                if (
                    caller_can_handle_requeue
                    and requeue_line_info
                    and requeue_line_info.lines_to_requeue
                ):
                    POGGER.debug("BOOOM-->break")
                    break
                assert not (requeue_line_info and requeue_line_info.lines_to_requeue)
                POGGER.debug(
                    "cob->process_link_reference_definition>>outer_processed>>$",
                    outer_processed,
                )
                POGGER.debug(
                    "cob->process_link_reference_definition>>did_complete_lrd>>$",
                    did_complete_lrd,
                )
                POGGER.debug(
                    "cob->process_link_reference_definition>>adjusted_tokens>>$",
                    adjusted_tokens,
                )
                assert not did_pause_lrd
            else:
                POGGER.debug(
                    "cob-rem>>$",
                    parser_state.token_document,
                )
                adjusted_tokens = TokenizedMarkdown.__remove_top_element_from_stack(
                    parser_state, was_close_forced
                )
                POGGER.debug(
                    "cob-rem<<$",
                    parser_state.token_document,
                )
                POGGER.debug("cob-adj<<$", adjusted_tokens)

            new_tokens.extend(adjusted_tokens)

        POGGER.debug("cob-end>>$", parser_state.token_stack)
        POGGER.debug("cob-end>>$", parser_state.token_document)
        POGGER.debug("cob-end>>new_tokens>>$", new_tokens)
        return new_tokens, requeue_line_info

    # pylint: enable=too-many-arguments,too-many-locals,too-many-statements, too-many-branches

    @staticmethod
    def __remove_top_element_from_stack(parser_state, was_forced):
        """
        Once it is decided that we need to remove the top element from the stack,
        make sure to do it uniformly.
        """

        new_tokens = []
        POGGER.debug("cob->top_element->$", parser_state.token_stack[-1])
        POGGER.debug("cob->was_forced->$", was_forced)
        extra_elements = []
        if parser_state.token_stack[-1].is_indented_code_block:
            extra_elements.extend(
                ContainerBlockProcessor.extract_markdown_tokens_back_to_blank_line(
                    parser_state, was_forced
                )
            )

        new_tokens.append(
            parser_state.token_stack[-1].generate_close_markdown_token_from_stack_token(
                was_forced=was_forced
            )
        )
        new_tokens.extend(extra_elements)
        del parser_state.token_stack[-1]
        return new_tokens

    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    @staticmethod
    def __handle_blank_line(
        parser_state,
        input_line,
        from_main_transform,
        position_marker=None,
        forced_close_until_index=None,
    ):
        """
        Handle the processing of a blank line.
        """

        if not from_main_transform:
            close_only_these_blocks = [ParagraphStackToken]
            do_include_block_quotes = False
        else:
            close_only_these_blocks = None
            do_include_block_quotes = True
        POGGER.debug("hbl>>from_main_transform>>$", from_main_transform)
        POGGER.debug("hbl>>close_only_these_blocks>>$", close_only_these_blocks)
        POGGER.debug("hbl>>do_include_block_quotes>>$", do_include_block_quotes)
        POGGER.debug("hbl>>forced_close_until_index>>$", forced_close_until_index)

        non_whitespace_index, extracted_whitespace = ParserHelper.extract_whitespace(
            input_line, 0
        )

        is_processing_list, in_index = LeafBlockProcessor.check_for_list_in_process(
            parser_state
        )
        POGGER.debug(
            "hbl>>is_processing_list>>$>>in_index>>$>>last_stack>>$",
            is_processing_list,
            in_index,
            parser_state.token_stack[-1],
        )

        requeue_line_info = None
        new_tokens = None
        force_default_handling = False
        if parser_state.token_stack[-1].was_link_definition_started:
            POGGER.debug(
                "hbl>>process_link_reference_definition>>stopping link definition"
            )
            empty_position_marker = PositionMarker(-1, 0, "")
            (
                _,
                _,
                did_pause_lrd,
                requeue_line_info,
                new_tokens,
            ) = LinkReferenceDefinitionHelper.process_link_reference_definition(
                parser_state, empty_position_marker, "", "", "", 0, 0
            )
            assert not did_pause_lrd
            force_default_handling = True
        elif parser_state.token_stack[-1].is_code_block:
            stack_bq_count = parser_state.count_of_block_quotes_on_stack()
            if stack_bq_count:
                POGGER.debug("hbl>>code block within block quote")
            else:
                POGGER.debug("hbl>>code block")
                new_tokens = []
        elif parser_state.token_stack[-1].is_html_block:
            POGGER.debug("hbl>>check_blank_html_block_end")
            new_tokens = HtmlHelper.check_blank_html_block_end(parser_state)
        elif (
            is_processing_list
            and parser_state.token_document[-1].is_blank_line
            and parser_state.token_document[-2].is_list_start
        ):
            POGGER.debug("hbl>>double blank in list")
            new_tokens, _ = TokenizedMarkdown.__close_open_blocks(
                parser_state, until_this_index=in_index, include_lists=True
            )
        elif forced_close_until_index:
            POGGER.debug("hbl>>forced_close_until_index")
            new_tokens, _ = TokenizedMarkdown.__close_open_blocks(
                parser_state,
                until_this_index=forced_close_until_index,
                include_lists=True,
                include_block_quotes=True,
            )

        if from_main_transform:
            POGGER.debug("hbl>>__handle_blank_line_in_block_quote")
            TokenizedMarkdown.__handle_blank_line_in_block_quote(parser_state)

        if force_default_handling or new_tokens is None:
            POGGER.debug("hbl>>default blank handling-->cob")
            n_tokens, _ = TokenizedMarkdown.__close_open_blocks(
                parser_state,
                only_these_blocks=close_only_these_blocks,
                include_block_quotes=do_include_block_quotes,
                was_forced=True,
            )
            if new_tokens:
                new_tokens.extend(n_tokens)
            else:
                new_tokens = n_tokens

        POGGER.debug("hbl>>new_tokens>>$", new_tokens)
        assert non_whitespace_index == len(input_line)
        if not (requeue_line_info and requeue_line_info.force_ignore_first_as_lrd):
            new_tokens.append(
                BlankLineMarkdownToken(extracted_whitespace, position_marker)
            )
        POGGER.debug("hbl>>new_tokens>>$", new_tokens)

        return new_tokens, requeue_line_info

    # pylint: enable=too-many-locals, too-many-branches, too-many-statements

    @staticmethod
    def __handle_blank_line_in_block_quote(parser_state):

        stack_index = parser_state.find_last_container_on_stack()
        POGGER.debug(
            "blank>>bq_start>>$",
            parser_state.token_stack[stack_index],
        )
        if stack_index > 0 and parser_state.token_stack[stack_index].is_block_quote:
            parser_state.token_stack[
                stack_index
            ].matching_markdown_token.add_leading_spaces("")

    def __process_header_if_present(self, token_to_use, line_number, requeue):

        if self.__front_matter_enabled:
            (
                token_to_use,
                line_number,
                requeue,
            ) = FrontMatterExtension.process_header_if_present(
                token_to_use,
                line_number,
                requeue,
                self.source_provider,
                self.tokenized_document,
            )
        return token_to_use, line_number, requeue


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that looks for hard tabs in the files.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd013(Plugin):
    """
    Class to implement a plugin that looks for hard tabs in the files.
    """

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            # line_length
            plugin_name="line-length",
            plugin_id="MD013",
            plugin_enabled_by_default=False,
            plugin_description="Line length",
            plugin_version="0.0.0",
            plugin_interface_version=1,
        )  # https://github.com/DavidAnson/markdownlint/blob/master/doc/Rules.md#md013---line-length
        # Parameters: line_length, heading_line_length, code_block_line_length, code_blocks, tables, headings, headers, strict (number; default 80, boolean; default true)



"""
Module to implement a plugin that looks for hard tabs in the files.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd030(Plugin):
    """
    Class to implement a plugin that looks for hard tabs in the files.
    """

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            # ol, ul, whitespace
            plugin_name="list-marker-space",
            plugin_id="MD030",
            plugin_enabled_by_default=False,
            plugin_description="Spaces after list markers",
            plugin_version="0.0.0",
            plugin_interface_version=1,
        )  # https://github.com/DavidAnson/markdownlint/blob/master/doc/Rules.md#md030---spaces-after-list-markers
        # Parameters: ul_single, ol_single, ul_multi, ol_multi (number; default 1)



"""
Module to implement a plugin that looks for text in a paragraph where a line starts
with what could be an atx heading, except there is no spaces between the hashes and
the text of the heading.
"""
import re

from pymarkdown.plugin_manager import Plugin, PluginDetails


class StartOfLineTokenParser:
    """
    Token parser that activates a check only on valid newlines within
    a paragraph.
    """

    def __init__(self):
        self.__last_paragraph_token = None
        self.__paragraph_index = None
        self.__first_line_after_other_token = None
        self.__paragraph_column_number = None
        self.__inside_of_link = None
        self.__first_line_after_hard_break = None
        self.__delayed_line = None

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_paragraph_token = None
        self.__paragraph_index = None
        self.__first_line_after_other_token = None
        self.__paragraph_column_number = None

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_paragraph:
            self.__next_token_paragraph_start(token)
        elif token.is_paragraph_end:
            self.__next_token_paragraph_end()
        elif self.__last_paragraph_token:
            if self.__inside_of_link:
                if token.is_inline_link_end:
                    self.__inside_of_link = False
            elif token.is_text:
                self.__next_token_paragraph_text_inline(token, context)
            else:
                self.__next_token_paragraph_non_text_inline(token)

    def __next_token_paragraph_start(self, token):
        self.__last_paragraph_token = token
        self.__paragraph_index = 0
        self.__first_line_after_other_token = True
        self.__first_line_after_hard_break = False
        self.__inside_of_link = False
        self.__paragraph_column_number = token.column_number
        self.__delayed_line = None

    def __next_token_paragraph_end(self):
        if self.__delayed_line:
            self.check_start_of_line(
                self.__delayed_line[0],
                self.__delayed_line[1],
                self.__delayed_line[2],
                self.__delayed_line[3],
                self.__delayed_line[4],
            )
            self.__delayed_line = None
        self.__last_paragraph_token = None

    def __next_token_paragraph_non_text_inline(self, token):
        if token.is_inline_code_span:
            self.__delayed_line = None
            self.__paragraph_index += (
                token.leading_whitespace.count("\n")
                + token.span_text.count("\n")
                + token.trailing_whitespace.count("\n")
            )
        elif token.is_inline_raw_html:
            self.__delayed_line = None
            self.__paragraph_index += token.raw_tag.count("\n")
        elif token.is_inline_image or token.is_inline_link:
            self.__delayed_line = None
            self.__paragraph_index += token.text_from_blocks.count("\n")
            if token.label_type == "inline":
                self.__paragraph_index += token.before_link_whitespace.count("\n")
                self.__paragraph_index += token.before_title_whitespace.count("\n")
                self.__paragraph_index += token.after_title_whitespace.count("\n")
                self.__paragraph_index += token.active_link_title.count("\n")
            if token.label_type == "full":
                self.__paragraph_index += token.ex_label.count("\n")
            self.__inside_of_link = token.is_inline_link
        elif token.is_inline_hard_break:
            self.__delayed_line = None
            self.__paragraph_index += 1
            self.__first_line_after_hard_break = True
        else:
            assert (
                token.is_inline_emphasis
                or token.is_inline_emphasis_end
                or token.is_inline_autolink
            )
            self.__delayed_line = None

    def __next_token_paragraph_text_inline(self, token, context):
        split_whitespace = self.__last_paragraph_token.extracted_whitespace.split("\n")
        split_text = token.token_text.split("\n")

        for split_index, next_text in enumerate(split_text):
            combined_text = (
                split_whitespace[split_index + self.__paragraph_index] + next_text
            )
            if (
                self.__first_line_after_hard_break
                or self.__first_line_after_other_token
            ) or split_index:
                if self.__first_line_after_other_token:
                    adjusted_column_number = self.__paragraph_column_number
                else:
                    adjusted_column_number = self.__paragraph_column_number + len(
                        split_whitespace[split_index + self.__paragraph_index]
                    )
                # pylint: disable=invalid-unary-operand-type
                if split_index == len(split_text) - 1:
                    self.__delayed_line = (
                        combined_text,
                        context,
                        token,
                        split_index,
                        -adjusted_column_number,
                    )
                else:
                    self.check_start_of_line(
                        combined_text,
                        context,
                        token,
                        split_index,
                        -adjusted_column_number,
                    )
                # pylint: enable=invalid-unary-operand-type
            self.__first_line_after_other_token = False
            self.__first_line_after_hard_break = False
        self.__paragraph_index += token.token_text.count("\n")

    # pylint: disable=too-many-arguments
    def check_start_of_line(
        self, combined_text, context, token, line_number_delta, column_number_delta
    ):
        """
        Check for a pattern at the start of the line.
        """

    # pylint: enable=too-many-arguments


class MyStartOfLineTokenParser(StartOfLineTokenParser):
    """
    Local implementation of the token parser.
    """

    def __init__(self, owner):
        super().__init__()
        self.__owner = owner

    # pylint: disable=too-many-arguments
    def check_start_of_line(
        self, combined_text, context, token, line_number_delta, column_number_delta
    ):
        """
        Check for a pattern at the start of the line.

        Note that the logic here is a bit weird for the `\\S` at the end of the
        regular expression.  If everything else is good except for that `\\S`,
        which is equivalent to any whitespace character, and it was a space character,
        it would match the Atx code in the leaf processor and be an Atx Heading token
        already.  Hence, what is left is any other whitespace character which is not
        the space character, which should still trigger the rule.
        """
        if re.search(r"^\s{0,3}#{1,6}\S", combined_text) and not re.search(
            r"#\s*$", combined_text
        ):
            self.__owner.report_next_token_error(
                context,
                token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
            )

    # pylint: enable=too-many-arguments


class RuleMd018(Plugin):
    """
    Class to implement a plugin that looks for text in a paragraph where a line starts
    with what could be an atx heading, except there is no spaces between the hashes and
    the text of the heading.
    """

    def __init__(self):
        super().__init__()
        self.__token_parser = MyStartOfLineTokenParser(self)

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-missing-space-atx",
            plugin_id="MD018",
            plugin_enabled_by_default=True,
            plugin_description="No space present after the hash character on a possible Atx Heading.",
            plugin_version="0.5.0",
            plugin_interface_version=1,
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__token_parser.starting_new_file()

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        self.__token_parser.next_token(context, token)



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert

# pylint: disable=too-many-lines


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_hb():
    """
    Test case:  Ordered list newline html block
    was:        test_list_blocks_256k
    """

    # Arrange
    source_markdown = """1.
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[end-olist:::True]",
        "[html-block(2,1)]",
        "[text(2,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li></li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_all_i3_hb():
    """
    Test case:  Ordered list newline (all indented) html block
    """

    # Arrange
    source_markdown = """1.
   <script>
   foo
   </script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   \n   ]",
        "[BLANK(1,3):]",
        "[html-block(2,4)]",
        "[text(2,4):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<script>
foo
</script>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i2_hb():
    """
    Test case:  Ordered list newline indent of 2 html block
    was:        test_list_blocks_256kxa
    """

    # Arrange
    source_markdown = """1.
  <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[end-olist:::True]",
        "[html-block(2,1)]",
        "[text(2,3):<script>\nfoo\n</script>:  ]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li></li>
</ol>
  <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_hb():
    """
    Test case:  Ordered list newline indent of 3 html block
    was:        test_list_blocks_256kxb
    """

    # Arrange
    source_markdown = """1.
   <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[BLANK(1,3):]",
        "[html-block(2,4)]",
        "[text(2,4):<script>:]",
        "[end-html-block:::True]",
        "[end-olist:::True]",
        "[para(3,1):\n]",
        "[text(3,1):foo\n::\n]",
        "[raw-html(4,1):/script]",
        "[end-para:::True]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<script>
</li>
</ol>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_hb():
    """
    Test case:  Ordered list text newline html block
    was:        test_list_blocks_256ka
    """

    # Arrange
    source_markdown = """1.  abc
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:4:]",
        "[para(1,5):]",
        "[text(1,5):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[html-block(2,1)]",
        "[text(2,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_all_i4_hb():
    """
    Test case:  Ordered list text newline html block
    """

    # Arrange
    source_markdown = """1.  abc
    <script>
    foo
    </script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:4::    \n    \n    ]",
        "[para(1,5):]",
        "[text(1,5):abc:]",
        "[end-para:::False]",
        "[html-block(2,5)]",
        "[text(2,5):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>abc
<script>
foo
</script>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_hb():
    """
    Test case:  Ordered list text newline indent of 3 html block
    was:        test_list_blocks_256kaa
    """

    # Arrange
    source_markdown = """1.  abc
   <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:4:]",
        "[para(1,5):]",
        "[text(1,5):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[html-block(2,1)]",
        "[text(2,4):<script>\nfoo\n</script>:   ]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc</li>
</ol>
   <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i4_hb():
    """
    Test case:  Ordered list text newline indent of 4 html block
    was:        test_list_blocks_256kab
    """

    # Arrange
    source_markdown = """1.  abc
    <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:4::    ]",
        "[para(1,5):]",
        "[text(1,5):abc:]",
        "[end-para:::False]",
        "[html-block(2,5)]",
        "[text(2,5):<script>:]",
        "[end-html-block:::True]",
        "[end-olist:::True]",
        "[para(3,1):\n]",
        "[text(3,1):foo\n::\n]",
        "[raw-html(4,1):/script]",
        "[end-para:::True]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<script>
</li>
</ol>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_ol_nl_hb():
    """
    Test case:  Ordered list x2 newline html block
    """

    # Arrange
    source_markdown = """1. 1.
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   ]",
        "[BLANK(1,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(2,1)]",
        "[text(2,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_ol_nl_all_i6_hb():
    """
    Test case:  Ordered list x2 newline (all indented) html block
    """

    # Arrange
    source_markdown = """1. 1.
      <script>
      foo
      </script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   :      \n      \n      ]",
        "[BLANK(1,6):]",
        "[html-block(2,7)]",
        "[text(2,7):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>
<script>
foo
</script>
</li>
</ol>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_ol_t_nl_hb():
    """
    Test case:  Ordered list x2 text newline html block
    was:        test_list_blocks_256kb
    """

    # Arrange
    source_markdown = """1. 1. abc
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   ]",
        "[para(1,7):]",
        "[text(1,7):abc:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(2,1)]",
        "[text(2,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc</li>
</ol>
</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_ol_t_nl_all_i6_hb():
    """
    Test case:  Ordered list x2 text newline html block
    """

    # Arrange
    source_markdown = """1. 1. abc
      <script>
      foo
      </script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   :      \n      \n      ]",
        "[para(1,7):]",
        "[text(1,7):abc:]",
        "[end-para:::False]",
        "[html-block(2,7)]",
        "[text(2,7):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc
<script>
foo
</script>
</li>
</ol>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_ol_nl_hb():
    """
    Test case:  Ordered list newline indent of 3 ordered list newline html block
    """

    # Arrange
    source_markdown = """1.
   1.
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_ol_nl_hb():
    """
    Test case:  Ordered list text newline indent of 3 ordered list newline html block
    """

    # Arrange
    source_markdown = """1. abc
   1.
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\n1.::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ul_t_nl_i2_ul_nl_hb():
    """
    Test case:  Unordered list text newline indent of 2 unordered list newline html block
    """

    # Arrange
    source_markdown = """- abc
  -
<script>
foo
</script>
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[end-ulist:::True]",
        "[html-block(3,1)]",
        "[text(3,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
</li>
</ul>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ul_t_nl_i2_ulb_nl_hb():
    """
    Test case:  Unordered list text newline indent of 2 unordered list (b) newline html block
    """

    # Arrange
    source_markdown = """- abc
  *
<script>
foo
</script>
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[html-block(3,1)]",
        "[text(3,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*</li>
</ul>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_ol_t_nl_hb():
    """
    Test case:  Ordered list newline indent of 3 ordered list text newline html block
    """

    # Arrange
    source_markdown = """1.
   1. def
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):def:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>def</li>
</ol>
</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_ol_t_nl_hb():
    """
    Test case:  Ordered list text newline indent of 3 ordered list text newline html block
    was:        test_list_blocks_256kc
    """

    # Arrange
    source_markdown = """1. abc
   1. def
<script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):def:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,1):<script>\nfoo\n</script>:]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>def</li>
</ol>
</li>
</ol>
<script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_ol_nl_i2_hb():
    """
    Test case:  Ordered list newline indent of 3 ordered list newline indent of 2 html block
    """

    # Arrange
    source_markdown = """1.
   1.
  <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,3):<script>\nfoo\n</script>:  ]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
  <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_ol_nl_i2_hb():
    """
    Test case:  Ordered list text newline indent of 3 ordered list newline indent of 2 html block
    """

    # Arrange
    source_markdown = """1. abc
   1.
  <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\n1.::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,3):<script>\nfoo\n</script>:  ]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.</li>
</ol>
  <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ul_t_nl_i2_ul_nl_i1_hb():
    """
    Test case:  Unordered list text newline indent of 2 unordered list newline indent of 1 html block
    """

    # Arrange
    source_markdown = """- abc
  -
 <script>
foo
</script>
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[end-ulist:::True]",
        "[html-block(3,1)]",
        "[text(3,2):<script>\nfoo\n</script>: ]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
</li>
</ul>
 <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ul_t_nl_i2_ulb_nl_i1_hb():
    """
    Test case:  Unordered list text newline indent of 2 unordered list (b) newline indent of 1 html block
    """

    # Arrange
    source_markdown = """- abc
  *
 <script>
foo
</script>
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[html-block(3,1)]",
        "[text(3,2):<script>\nfoo\n</script>: ]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*</li>
</ul>
 <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_ol_t_nl_i2_hb():
    """
    Test case:  Ordered list newline indent of 3 ordered list text newline indent of 2 html block
    """

    # Arrange
    source_markdown = """1.
   1. def
  <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):def:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,3):<script>\nfoo\n</script>:  ]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>def</li>
</ol>
</li>
</ol>
  <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_ol_t_nl_i2_hb():
    """
    Test case:  Ordered list text newline indent of 3 ordered list text newline indent of 2 html block
    was:        test_list_blocks_256kd
    """

    # Arrange
    source_markdown = """1. abc
   1. def
  <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):def:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[html-block(3,1)]",
        "[text(3,3):<script>\nfoo\n</script>:  ]",
        "[end-html-block:::False]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>def</li>
</ol>
</li>
</ol>
  <script>
foo
</script>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_ol_nl_i3_hb():
    """
    Test case:  Ordered list newline indent of 3 ordered list newline indent of 2 html block
    """

    # Arrange
    source_markdown = """1.
   1.
   <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[html-block(3,4)]",
        "[text(3,4):<script>:]",
        "[end-html-block:::True]",
        "[end-olist:::True]",
        "[para(4,1):\n]",
        "[text(4,1):foo\n::\n]",
        "[raw-html(5,1):/script]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
<script>
</li>
</ol>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_ol_nl_i3_hb():
    """
    Test case:  Ordered list text newline indent of 3 ordered list newline indent of 2 html block
    """

    # Arrange
    source_markdown = """1. abc
   1.
   <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\n1.::\n]",
        "[end-para:::False]",
        "[html-block(3,4)]",
        "[text(3,4):<script>:]",
        "[end-html-block:::True]",
        "[end-olist:::True]",
        "[para(4,1):\n]",
        "[text(4,1):foo\n::\n]",
        "[raw-html(5,1):/script]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.
<script>
</li>
</ol>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ul_t_nl_i2_ul_nl_i2_hb():
    """
    Test case:  Unordered list text newline indent of 2 unordered list newline indent of 2 html block
    """

    # Arrange
    source_markdown = """- abc
  -
  <script>
foo
</script>
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[html-block(3,3)]",
        "[text(3,3):<script>:]",
        "[end-html-block:::True]",
        "[end-ulist:::True]",
        "[para(4,1):\n]",
        "[text(4,1):foo\n::\n]",
        "[raw-html(5,1):/script]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
<script>
</li>
</ul>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ul_t_nl_i2_ulb_nl_i2_hb():
    """
    Test case:  Unordered list text newline indent of 2 unordered list (b) newline indent of 2 html block
    """

    # Arrange
    source_markdown = """- abc
  *
  <script>
foo
</script>
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[para(1,3):\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[end-para:::False]",
        "[html-block(3,3)]",
        "[text(3,3):<script>:]",
        "[end-html-block:::True]",
        "[end-ulist:::True]",
        "[para(4,1):\n]",
        "[text(4,1):foo\n::\n]",
        "[raw-html(5,1):/script]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*
<script>
</li>
</ul>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_nl_i3_ol_t_nl_i3_hb():
    """
    Test case:  Ordered list newline indent of 3 ordered list text newline indent of 2 html block
    """

    # Arrange
    source_markdown = """1.
   1. def
   <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   :   ]",
        "[para(2,7):]",
        "[text(2,7):def:]",
        "[end-para:::False]",
        "[end-olist:::True]",
        "[html-block(3,4)]",
        "[text(3,4):<script>:]",
        "[end-html-block:::True]",
        "[end-olist:::True]",
        "[para(4,1):\n]",
        "[text(4,1):foo\n::\n]",
        "[raw-html(5,1):/script]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>def</li>
</ol>
<script>
</li>
</ol>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hb_ol_t_nl_i3_ol_t_nl_i3_hb():
    """
    Test case:  Ordered list text newline indent of 3 ordered list text newline indent of 2 html block
    was:        test_list_blocks_256ke
    """

    # Arrange
    source_markdown = """1. abc
   1. def
   <script>
foo
</script>
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   :   ]",
        "[para(2,7):]",
        "[text(2,7):def:]",
        "[end-para:::False]",
        "[end-olist:::True]",
        "[html-block(3,4)]",
        "[text(3,4):<script>:]",
        "[end-html-block:::True]",
        "[end-olist:::True]",
        "[para(4,1):\n]",
        "[text(4,1):foo\n::\n]",
        "[raw-html(5,1):/script]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>def</li>
</ol>
<script>
</li>
</ol>
<p>foo
</script></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide a tokenization of a markdown-encoded string.
"""


# pylint: disable=too-few-public-methods
class InMemorySourceProvider:
    """
    Class to provide for a source provider that is totally within memory.
    """

    def __init__(self, source_text):
        self.next_token = source_text.split("\n", 1)

    def get_next_line(self):
        """
        Get the next line from the source provider.
        """
        token_to_use = None
        if self.next_token:
            if len(self.next_token) == 2:
                token_to_use = self.next_token[0]
                self.next_token = self.next_token[1].split("\n", 1)
            else:
                assert self.next_token
                token_to_use = self.next_token[0]
                self.next_token = None
        return token_to_use


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class FileSourceProvider:
    """
    Class to provide for a source provider that is on media as a file.
    """

    def __init__(self, file_to_open):
        with open(file_to_open, encoding="utf-8") as file_to_parse:
            file_as_lines = file_to_parse.readlines()

        self.read_lines = []
        self.read_index = 0
        did_line_end_in_newline = True
        for next_line in file_as_lines:
            did_line_end_in_newline = next_line.endswith("\n")
            if did_line_end_in_newline:
                next_line = next_line[0:-1]
            self.read_lines.append(next_line)

        if did_line_end_in_newline:
            self.read_lines.append("")

    def get_next_line(self):
        """
        Get the next line from the source provider.
        """
        token_to_use = None
        if self.read_index < len(self.read_lines):
            token_to_use = self.read_lines[self.read_index]
            self.read_index += 1
        return token_to_use


# pylint: enable=too-few-public-methods



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_paragraph_series_j_l_bh_s_t():
    """
    Test case:  Inline link containing backslash hard line break in label, followed by text
    was:        test_paragraph_extra_h8
    """

    # Arrange
    source_markdown = """a[foo\\\ncom](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::foo\\\ncom:False:":: :]',
        "[text(1,3):foo:]",
        "[hard-break(1,6):\\:\n]",
        "[text(2,1):com:]",
        "[end-link::]",
        "[text(2,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">foo<br />
com</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_bh_s_t():
    """
    Test case:  Inline image containing backslash hard line break in label, followed by text
    was:        test_paragraph_extra_h8a
    """

    # Arrange
    source_markdown = """a![foo\\\ncom](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:foo\ncom::::foo\\\ncom:False:":: :]',
        "[text(2,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="foo
com" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_sh_s_t():
    """
    Test case:  Inline image containing space hard line break in label, followed by text
    was:        test_paragraph_extra_h8b
    """

    # Arrange
    source_markdown = """a[foo  \ncom](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::foo  \ncom:False:":: :]',
        "[text(1,3):foo:]",
        "[hard-break(1,6):  :\n]",
        "[text(2,1):com:]",
        "[end-link::]",
        "[text(2,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">foo<br />
com</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_sh_s_t():
    """
    Test case:  Inline image containing space hard line break in label, followed by text
    was:        test_paragraph_extra_h8c
    """

    # Arrange
    source_markdown = """a![foo  \ncom](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:foo\ncom::::foo  \ncom:False:":: :]',
        "[text(2,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="foo
com" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_rh_s_t():
    """
    Test case:  Inline link containing split raw html in label
    was:        test_paragraph_extra_c7, changed split text at end to text
    """

    # Arrange
    source_markdown = """a[li<de\nfg>nk](/url)a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/url:::::li<de\nfg>nk:False::::]",
        "[text(1,3):li:]",
        "[raw-html(1,5):de\nfg]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,13):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url">li<de
fg>nk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_rh_s_t():
    """
    Test case:  Inline image containing split raw html in label
    was:        test_paragraph_extra_c8, changed split text at end to text
    """

    # Arrange
    source_markdown = """a![li<de\nfg>nk](/url)a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):inline:/url::li<de\nfg>nk::::li<de\nfg>nk:False::::]",
        "[text(2,13):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li<de
fg>nk" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_em_s_t():
    """
    Test case:  Inline link containing split emphasis in label
    was:        test_paragraph_extra_e1
    """

    # Arrange
    source_markdown = """a[a*li\nnk*a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::a*li\nnk*a:False:":: :]',
        "[text(1,3):a:]",
        "[emphasis(1,4):1:*]",
        "[text(1,5):li\nnk::\n]",
        "[end-emphasis(2,3)::]",
        "[text(2,4):a:]",
        "[end-link::]",
        "[text(2,20):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">a<em>li\nnk</em>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_em_s_t():
    """
    Test case:  Inline image containing split emphasis in label
    was:        test_paragraph_extra_e2
    """

    # Arrange
    source_markdown = """a![a*li\nnk*a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:ali\nnka::::a*li\nnk*a:False:":: :]',
        "[text(2,20):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="ali\nnka" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_em_t():
    """
    Test case:  Inline image containing emphasis in label
    was:        test_paragraph_extra_d9
    """

    # Arrange
    source_markdown = """a[*link*](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::*link*:False:":: :]',
        "[emphasis(1,3):1:*]",
        "[text(1,4):link:]",
        "[end-emphasis(1,8)::]",
        "[end-link::]",
        "[text(1,24):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title"><em>link</em></a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_em_t():
    """
    Test case:  Inline link containing emphasis in label
    was:        test_paragraph_extra_e0
    """

    # Arrange
    source_markdown = """a![*link*](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:link::::*link*:False:":: :]',
        "[text(1,25):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="link" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_ua_t():
    """
    Test case:  Inline link containing URI autolink in label
    was:        test_paragraph_extra_h6
    """

    # Arrange
    source_markdown = """a[a<http://google.com>a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::a<http://google.com>a:False:":: :]',
        "[text(1,3):a:]",
        "[uri-autolink(1,4):http://google.com]",
        "[text(1,23):a:]",
        "[end-link::]",
        "[text(1,39):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">a<a href="http://google.com">http://google.com</a>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_ua_t():
    """
    Test case:  Inline image containing URI autolink in label
    was:        test_paragraph_extra_h6a
    """

    # Arrange
    source_markdown = """a![a<http://google.com>a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:ahttp://google.coma::::a<http://google.com>a:False:":: :]',
        "[text(1,40):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>a<img src="/uri" alt="ahttp://google.coma" title="title" />a</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_ea_t():
    """
    Test case:  Inline link containing email autolink in label
    was:        test_paragraph_extra_h7
    """

    # Arrange
    source_markdown = """a[a<foo@r.com>a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::a<foo@r.com>a:False:":: :]',
        "[text(1,3):a:]",
        "[email-autolink(1,4):foo@r.com]",
        "[text(1,15):a:]",
        "[end-link::]",
        "[text(1,31):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">a<a href="mailto:foo@r.com">foo@r.com</a>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_ea_t():
    """
    Test case:  Inline image containing email autolink in label
    was:        test_paragraph_extra_h7a
    """

    # Arrange
    source_markdown = """a![a<foo@r.com>a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:afoo@r.coma::::a<foo@r.com>a:False:":: :]',
        "[text(1,32):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="afoo@r.coma" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_cs_t():
    """
    Test case:  Inline link containing code span in label
    was:        test_paragraph_extra_h4
    """

    # Arrange
    source_markdown = """a[a`li nk`a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::a`li nk`a:False:":: :]',
        "[text(1,3):a:]",
        "[icode-span(1,4):li nk:`::]",
        "[text(1,11):a:]",
        "[end-link::]",
        "[text(1,27):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">a<code>li nk</code>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_cs_t():
    """
    Test case:  Inline image containing code span in label
    was:        test_paragraph_extra_h4a
    """

    # Arrange
    source_markdown = """a![a`li nk`a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:ali nka::::a`li nk`a:False:":: :]',
        "[text(1,28):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="ali nka" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_rh_t():
    """
    Test case:  Inline link containing code span in label
    was:        test_paragraph_extra_h5
    """

    # Arrange
    source_markdown = """a[a<li nk>a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::a<li nk>a:False:":: :]',
        "[text(1,3):a:]",
        "[raw-html(1,4):li nk]",
        "[text(1,11):a:]",
        "[end-link::]",
        "[text(1,27):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">a<li nk>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_rh_t():
    """
    Test case:  Inline image containing code span in label
    was:        test_paragraph_extra_h5a
    """

    # Arrange
    source_markdown = """a![a<li nk>a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:a<li nk>a::::a<li nk>a:False:":: :]',
        "[text(1,28):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="a<li nk>a" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_cs_s_t():
    """
    Test case:  Inline link containing split code span in label
    """

    # Arrange
    source_markdown = """a[a`li\nnk`a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::a`li\nnk`a:False:":: :]',
        "[text(1,3):a:]",
        "[icode-span(1,4):li\a\n\a \ank:`::]",
        "[text(2,4):a:]",
        "[end-link::]",
        "[text(2,20):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">a<code>li nk</code>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_cs_s_t():
    """
    Test case:  Inline image containing split code span in label
    """

    # Arrange
    source_markdown = """a![a`li\nnk`a](/uri "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:ali nka::::a`li\nnk`a:False:":: :]',
        "[text(2,20):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="ali nka" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_nt_t_s_t():
    """
    Test case:  Inline link with no title containing split text in label
    was:        test_paragraph_extra_a2
    """

    # Arrange
    source_markdown = """a![fo
o](</my url>)a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):inline:/my%20url::fo\no:/my url:::fo\no:True::::]",
        "[text(2,14):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/my%20url" alt="fo
o" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_nt_t_s_t():
    """
    Test case:  Inline image with no title containing split text in label
    was:        test_paragraph_extra_a2a
    """

    # Arrange
    source_markdown = """a[fo
o](</my url>)a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/my%20url::/my url:::fo\no:True::::]",
        "[text(1,3):fo\no::\n]",
        "[end-link::]",
        "[text(2,14):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/my%20url">fo
o</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_s_t():
    """
    Test case:  Inline link containing split text in label
    """

    # Arrange
    source_markdown = """a![fo
o](</my url> "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/my%20url:title:fo\no:/my url:::fo\no:True:":: :]',
        "[text(2,22):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/my%20url" alt="fo
o" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_s_t():
    """
    Test case:  Inline image containing split text in label
    """

    # Arrange
    source_markdown = """a[fo
o](</my url> "title")a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/my%20url:title:/my url:::fo\no:True:":: :]',
        "[text(1,3):fo\no::\n]",
        "[end-link::]",
        "[text(2,22):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/my%20url" title="title">fo
o</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_nt_t_t():
    """
    Test case:  Inline image with no title containing text in label
    """

    # Arrange
    source_markdown = """a![foo](</my url>)a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[image(1,2):inline:/my%20url::foo:/my url:::foo:True::::]",
        "[text(1,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/my%20url" alt="foo" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_nt_t_t():
    """
    Test case:  Inline link with no title containing text in label
    """

    # Arrange
    source_markdown = """a[foo](</my url>)a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/my%20url::/my url:::foo:True::::]",
        "[text(1,3):foo:]",
        "[end-link::]",
        "[text(1,18):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/my%20url">foo</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_t():
    """
    Test case:  Inline link containing text in label
    """

    # Arrange
    source_markdown = """a![foo](</my url> "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/my%20url:title:foo:/my url:::foo:True:":: :]',
        "[text(1,27):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/my%20url" alt="foo" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_t():
    """
    Test case:  Inline image containing text in label
    """

    # Arrange
    source_markdown = """a[foo](</my url> "title")a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/my%20url:title:/my url:::foo:True:":: :]',
        "[text(1,3):foo:]",
        "[end-link::]",
        "[text(1,26):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/my%20url" title="title">foo</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_t_s():
    """
    Test case:  Inline link containing text in label and split text after
    """

    # Arrange
    source_markdown = """a[foo](/url "title")a\nb"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::foo:False:":: :]',
        "[text(1,3):foo:]",
        "[end-link::]",
        "[text(1,21):a\nb::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">foo</a>a
b</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_t_s():
    """
    Test case:  Inline image containing text in label and split text after
    """

    # Arrange
    source_markdown = """a![foo](/url "title")a\nb"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:foo::::foo:False:":: :]',
        "[text(1,22):a\nb::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="foo" title="title" />a
b</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_cs_s():
    """
    Test case:  Inline link with no title containing split text in label
    was:        test_paragraph_extra_c9, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a[link](/url "title")`a\nb`"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[icode-span(1,22):a\a\n\a \ab:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><code>a b</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_cs_s():
    """
    Test case:  Inline link with no title containing split text in label
    was:        test_paragraph_extra_d0, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a![link](/url "title")`a\nb`"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[icode-span(1,23):a\a\n\a \ab:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>a<img src="/url" alt="link" title="title" /><code>a b</code></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_cs():
    """
    Test case:  Inline link with text in label followed by code span
    """

    # Arrange
    source_markdown = """a[link](/url "title")`ab`"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[icode-span(1,22):ab:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><code>ab</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_cs():
    """
    Test case:  Inline image with text in label followed by code span
    """

    # Arrange
    source_markdown = """a![link](/url "title")`ab`"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[icode-span(1,23):ab:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>a<img src="/url" alt="link" title="title" /><code>ab</code></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_rh_s():
    """
    Test case:  Inline link containing text in label with split raw html after
    was:        test_paragraph_extra_d1, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a[link](/url "title")<a\nb>"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[raw-html(1,22):a\nb]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><a
b></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_rh_s():
    """
    Test case:  Inline image containing text in label with split raw html after
    was:        test_paragraph_extra_d2, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a![link](/url "title")<a\nb>"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[raw-html(1,23):a\nb]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" title="title" /><a
b></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_rh():
    """
    Test case:  Inline link containing text in label with raw html after
    was:        test_paragraph_extra_d1, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a[link](/url "title")<a b>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[raw-html(1,22):a b]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><a b></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_rh():
    """
    Test case:  Inline image containing text in label with split raw html after
    was:        test_paragraph_extra_d2, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a![link](/url "title")<a b>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[raw-html(1,23):a b]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" title="title" /><a b></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_em_s():
    """
    Test case:  Inline link containing text in label with split emphasis after
    was:        test_paragraph_extra_d3, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a[link](/url)*a\nb*"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/url:::::link:False::::]",
        "[text(1,3):link:]",
        "[end-link::]",
        "[emphasis(1,14):1:*]",
        "[text(1,15):a\nb::\n]",
        "[end-emphasis(2,2)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url">link</a><em>a
b</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_em_s():
    """
    Test case:  Inline image containing text in label with split emphasis after
    was:        test_paragraph_extra_d4, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a![link](/url)*a\nb*"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):inline:/url::link::::link:False::::]",
        "[emphasis(1,15):1:*]",
        "[text(1,16):a\nb::\n]",
        "[end-emphasis(2,2)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" /><em>a
b</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_em():
    """
    Test case:  Inline link containing text in label with emphasis after
    """

    # Arrange
    source_markdown = """a[link](/url)*a b*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/url:::::link:False::::]",
        "[text(1,3):link:]",
        "[end-link::]",
        "[emphasis(1,14):1:*]",
        "[text(1,15):a b:]",
        "[end-emphasis(1,18)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url">link</a><em>a b</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_em():
    """
    Test case:  Inline image containing text in label with emphasis after
    was:        test_paragraph_extra_d4, added title, removed split raw html in label
    """

    # Arrange
    source_markdown = """a![link](/url)*a b*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[image(1,2):inline:/url::link::::link:False::::]",
        "[emphasis(1,15):1:*]",
        "[text(1,16):a b:]",
        "[end-emphasis(1,19)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" /><em>a b</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_em_l_t_t():
    """
    Test case:  Inline link containing text in label surrounded by emphasis
    was:        test_paragraph_extra_d7
    """

    # Arrange
    source_markdown = """a*[link](/uri "title")*a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[text(1,2):*:]",
        '[link(1,3):inline:/uri:title::::link:False:":: :]',
        "[text(1,4):link:]",
        "[end-link::]",
        "[text(1,23):*:]",
        "[text(1,24):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a*<a href="/uri" title="title">link</a>*a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_em_i_t_t():
    """
    Test case:  Inline link containing text in label surrounded by emphasis
    was:        test_paragraph_extra_d8
    """

    # Arrange
    source_markdown = """a*![link](/uri "title")*a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[text(1,2):*:]",
        '[image(1,3):inline:/uri:title:link::::link:False:":: :]',
        "[text(1,24):*:]",
        "[text(1,25):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a*<img src="/uri" alt="link" title="title" />*a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_ua():
    """
    Test case:  Inline link containing text in label followed by url autolink
    """

    # Arrange
    source_markdown = """a[link](/url "title")<http://google.com>a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[uri-autolink(1,22):http://google.com]",
        "[text(1,41):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><a href="http://google.com">http://google.com</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_ua():
    """
    Test case:  Inline image containing text in label followed by url autolink
    """

    # Arrange
    source_markdown = """a![link](/url "title")<http://google.com>a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[uri-autolink(1,23):http://google.com]",
        "[text(1,42):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" title="title" /><a href="http://google.com">http://google.com</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_ea():
    """
    Test case:  Inline link containing text in label followed by email autolink
    """

    # Arrange
    source_markdown = """a[link](/url "title")<foo@r.com>a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[email-autolink(1,22):foo@r.com]",
        "[text(1,33):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><a href="mailto:foo@r.com">foo@r.com</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_ea():
    """
    Test case:  Inline image containing text in label followed by email autolink
    """

    # Arrange
    source_markdown = """a![link](/url "title")<foo@r.com>a"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[email-autolink(1,23):foo@r.com]",
        "[text(1,34):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" title="title" /><a href="mailto:foo@r.com">foo@r.com</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_bh():
    """
    Test case:  Inline link containing text in label followed by backslash hard break
    """

    # Arrange
    source_markdown = """a[link](/url "title")\\\na"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[hard-break(1,22):\\:\n]",
        "[text(2,1):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><br />
a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_bh():
    """
    Test case:  Inline image containing text in label followed by backslash hard break
    """

    # Arrange
    source_markdown = """a![link](/url "title")\\\na"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[hard-break(1,23):\\:\n]",
        "[text(2,1):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" title="title" /><br />
a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_l_t_sh():
    """
    Test case:  Inline link containing text in label followed by spaces hard break
    """

    # Arrange
    source_markdown = """a[link](/url "title")   \na"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/url:title::::link:False:":: :]',
        "[text(1,3):link:]",
        "[end-link::]",
        "[hard-break(1,22):   :\n]",
        "[text(2,1):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">link</a><br />
a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_j_i_t_sh():
    """
    Test case:  Inline image containing text in label followed by spaces hard break
    """

    # Arrange
    source_markdown = """a![link](/url "title")   \na"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/url:title:link::::link:False:":: :]',
        "[hard-break(1,23):   :\n]",
        "[text(2,1):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="link" title="title" /><br />
a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_406():
    """
    Test case 406:  Rule 8: This is not strong emphasis, because the closing delimiter is preceded by whitespace:
    """

    # Arrange
    source_markdown = """__foo bar __"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):__:]",
        "[text(1,3):foo bar :]",
        "[text(1,11):__:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>__foo bar __</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_407():
    """
    Test case 407:  This is not strong emphasis, because the second __ is preceded by punctuation and followed by an alphanumeric:
    """

    # Arrange
    source_markdown = """__(__foo)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):__:]",
        "[text(1,3):(:]",
        "[text(1,4):__:]",
        "[text(1,6):foo):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>__(__foo)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_408():
    """
    Test case 408:  The point of this restriction is more easily appreciated with this example:
    """

    # Arrange
    source_markdown = """_(__foo__)_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):(:]",
        "[emphasis(1,3):2:_]",
        "[text(1,5):foo:]",
        "[end-emphasis(1,8)::]",
        "[text(1,10):):]",
        "[end-emphasis(1,11)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>(<strong>foo</strong>)</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_409():
    """
    Test case 409:  (part 1) Intraword strong emphasis is forbidden with __:
    """

    # Arrange
    source_markdown = """__foo__bar"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):__:]",
        "[text(1,3):foo:]",
        "[text(1,6):__:]",
        "[text(1,8):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>__foo__bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_410():
    """
    Test case 410:  (part 2) Intraword strong emphasis is forbidden with __:
    """

    # Arrange
    source_markdown = """__пристаням__стремятся"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):__:]",
        "[text(1,3):пристаням:]",
        "[text(1,12):__:]",
        "[text(1,14):стремятся:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>__пристаням__стремятся</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_411():
    """
    Test case 411:  (part 3) Intraword strong emphasis is forbidden with __:
    """

    # Arrange
    source_markdown = """__foo__bar__baz__"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):2:_]",
        "[text(1,3):foo:]",
        "[text(1,6):__:]",
        "[text(1,8):bar:]",
        "[text(1,11):__:]",
        "[text(1,13):baz:]",
        "[end-emphasis(1,16)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><strong>foo__bar__baz</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_412():
    """
    Test case 412:  This is strong emphasis, even though the closing delimiter is both left- and right-flanking, because it is followed by punctuation:
    """

    # Arrange
    source_markdown = """__(bar)__."""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):2:_]",
        "[text(1,3):(bar):]",
        "[end-emphasis(1,8)::]",
        "[text(1,10):.:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><strong>(bar)</strong>.</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide tests related to the MD023 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest


@pytest.mark.rules
def test_md023_good_proper_indent_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD023 directory that has an atx heading that is not indented
    from the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/proper_indent_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_good_proper_indent_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD023 directory that has an setext heading that is not indented
    from the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/proper_indent_setext.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indent_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has an atx heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indent_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md023/improper_indent_atx.md:3:3: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indent_atx_in_list_item():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has an atx heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indent_atx_in_list_item.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indent_atx_in_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has an atx heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indent_atx_in_block_quote.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indent_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has a setext heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indent_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md023/improper_indent_setext.md:4:3: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext.md:9:3: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext.md:14:1: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext.md:22:1: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indent_setext_in_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has a setext heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indent_setext_in_block_quote.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md023/improper_indent_setext_in_block_quote.md:4:5: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext_in_block_quote.md:9:5: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext_in_block_quote.md:14:3: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext_in_block_quote.md:22:3: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indent_setext_in_list_item():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has a setext heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indent_setext_in_list_item.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md023/improper_indent_setext_in_list_item.md:4:5: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext_in_list_item.md:9:6: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
        + "test/resources/rules/md023/improper_indent_setext_in_list_item.md:22:3: "
        + "MD023: Headings must start at the beginning of the line. (heading-start-left, header-start-left)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md023_bad_improper_indented_atx_after_emphasis():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md023 directory that has a setext heading that is indented from
    the start of the line.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--enable-rules",
        "MD023",
        "scan",
        "test/resources/rules/md023/improper_indented_atx_after_emphasis.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to implement a plugin to ensure all files end with a blank line.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd047(Plugin):
    """
    Class to implement a plugin to ensure all files end with a blank line.
    """

    def __init__(self):
        super().__init__()
        self.__last_line = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="single-trailing-newline",
            plugin_id="MD047",
            plugin_enabled_by_default=True,
            plugin_description="Each file should end with a single newline character.",
            plugin_version="0.5.0",
            plugin_interface_version=1,
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_line = None

    def next_line(self, context, line):
        """
        Event that a new line is being processed.
        """
        self.__last_line = line

    def completed_file(self, context):
        """
        Event that the file being currently scanned is now completed.
        """
        if self.__last_line:
            self.report_next_line_error(context, len(self.__last_line), -1)
